////////////////////////////////////////////////////////////
// BrightnessContrastObject.h
//
// Created by Roman Ryltsov roman@alax.info
// 
// $Id: BrightnessContrastObject.h 16 2008-07-12 16:02:03Z John Doe $

#pragma once

#include <dshow.h>
#include <dmo.h>
#include <dmodshow.h>
#include <atltypes.h>
#include "DmoBrightnessContrastSample_i.h"

#pragma comment(lib, "strmiids.lib")
#pragma comment(lib, "dmoguids.lib")
#pragma comment(lib, "msdmo.lib")

#if defined(_WIN32_WCE) && !defined(_CE_DCOM) && !defined(_CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA)
#error "Single-threaded COM objects are not properly supported on Windows CE platform, such as the Windows Mobile platforms that do not include full DCOM support. Define _CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA to force ATL to support creating single-thread COM object's and allow use of it's single-threaded COM object implementations. The threading model in your rgs file was set to 'Free' as that is the only threading model supported in non DCOM Windows CE platforms."
#endif

////////////////////////////////////////////////////////////
// Helpers

inline VOID AtlCheck(HRESULT nResult)
{
	if(FAILED(nResult))
		AtlThrow(nResult);
}

inline VOID AtlCheck(BOOL bValue, HRESULT nResult)
{
	if(!bValue)
		AtlThrow(nResult);
}

inline VOID AtlCheck(const VOID* pvValue, HRESULT nResult)
{
	if(!pvValue)
		AtlThrow(nResult);
}

// Shorthand macros

#define _A ATLASSERT
#define _V(x) ATLVERIFY(SUCCEEDED(x))

#define __C AtlCheck
#define __D AtlCheck

#define E_UNNAMED MAKE_HRESULT(SEVERITY_ERROR, 'x', __LINE__)

#define _Z0 ATLTRACE2
#define _Z1 ATLTRACE2
#define _Z2 ATLTRACE2
#define _Z3 ATLTRACE2
#define _Z4 ATLTRACE2
#define _Z5 ATLTRACE2

#include "rodmo.h"

////////////////////////////////////////////////////////////
// CBrightnessContrastObject

#define FOURCC_YUY2 mmioFOURCC('Y', 'U', 'Y', '2')

class ATL_NO_VTABLE CBrightnessContrastObject :
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CBrightnessContrastObject, &CLSID_BrightnessContrastObject>,
	public IDispatchImpl<IBrightnessContrastObject>,
	public IMediaObject
{
	typedef CMediaObjectT::CInput CInput;
	typedef CMediaObjectT::CInputMediaBuffer CInputMediaBuffer;
	typedef CMediaObjectT::COutputMediaBuffer COutputMediaBuffer;

public:

//DECLARE_REGISTRY_RESOURCEID(IDR_BRIGHTNESSCONTRASTOBJECT)

DECLARE_PROTECT_FINAL_CONSTRUCT()

DECLARE_GET_CONTROLLING_UNKNOWN()

BEGIN_COM_MAP(CBrightnessContrastObject)
	COM_INTERFACE_ENTRY(IMediaObject)
	COM_INTERFACE_ENTRY(IBrightnessContrastObject)
	COM_INTERFACE_ENTRY(IDispatch)
END_COM_MAP()

private:
	mutable CComAutoCriticalSection m_DataCriticalSection;
	CDmoMediaType m_pInputMediaType;
	CDmoMediaType m_pOutputMediaType;
	CInput m_Input;
	LONG m_nBrightness;
	LONG m_nContrast;

	static CSize GetExtent(const BITMAPINFOHEADER* pBitmapInfoHeader) throw()
	{
		_A(pBitmapInfoHeader);
		return CSize(pBitmapInfoHeader->biWidth, (pBitmapInfoHeader->biHeight >= 0) ? pBitmapInfoHeader->biHeight : -pBitmapInfoHeader->biHeight);
	}
	static CSize GetExtent(const VIDEOINFOHEADER* pVideoInfoHeader) throw()
	{
		_A(pVideoInfoHeader);
		if(!IsRectEmpty(&pVideoInfoHeader->rcTarget))
			return reinterpret_cast<const CRect&>(pVideoInfoHeader->rcTarget).Size();
		return GetExtent(&pVideoInfoHeader->bmiHeader);
	}
	static CSize GetExtent(const DMO_MEDIA_TYPE* pMediaType) throw()
	{
		_A(pMediaType);
		_A(IsEqualGUID(pMediaType->formattype, FORMAT_VideoInfo));
		_A(pMediaType->cbFormat >= sizeof (VIDEOINFOHEADER));
		return GetExtent((const VIDEOINFOHEADER*) pMediaType->pbFormat);
	}
	static CDmoMediaType CreateMediaType(LONG nWidth, LONG nHeight)
	{
		_A(nWidth % 2 == 0);
		CDmoMediaType pMediaType;
		pMediaType.Allocate(sizeof (VIDEOINFOHEADER));
		_A(!pMediaType->pUnk && pMediaType->cbFormat == sizeof (VIDEOINFOHEADER) && pMediaType->pbFormat);
		ZeroMemory(pMediaType->pbFormat, pMediaType->cbFormat);
		VIDEOINFOHEADER* pVideoInfoHeader = (VIDEOINFOHEADER*) pMediaType->pbFormat;
		BITMAPINFOHEADER* pBitmapInfoHeader = &pVideoInfoHeader->bmiHeader;
		pBitmapInfoHeader->biSize = sizeof *pBitmapInfoHeader;
		pBitmapInfoHeader->biWidth = nWidth;
		pBitmapInfoHeader->biHeight = nHeight;
		pBitmapInfoHeader->biPlanes = 1;
		pBitmapInfoHeader->biBitCount = 16;
		pBitmapInfoHeader->biCompression = FOURCC_YUY2;
		pBitmapInfoHeader->biSizeImage = nHeight * (nWidth * 2);
		reinterpret_cast<CRect&>(pVideoInfoHeader->rcSource).SetRect(0, 0, pBitmapInfoHeader->biWidth, pBitmapInfoHeader->biHeight);
		reinterpret_cast<CRect&>(pVideoInfoHeader->rcTarget).SetRect(0, 0, pBitmapInfoHeader->biWidth, pBitmapInfoHeader->biHeight);
		pMediaType->majortype = MEDIATYPE_Video;
		pMediaType->subtype = MEDIASUBTYPE_YUY2;
		pMediaType->bFixedSizeSamples = TRUE;
		pMediaType->bTemporalCompression = FALSE;
		pMediaType->lSampleSize = pBitmapInfoHeader->biSizeImage;
		pMediaType->formattype = FORMAT_VideoInfo;
		return pMediaType;
	}
	static CDmoMediaType CreateMediaType(SIZE Extent)
	{
		return CreateMediaType(Extent.cx, Extent.cy);
	}
	static BOOL IsValidMediaType(const CDmoMediaType& pMediaType) throw()
	{
		if(pMediaType)
		{
			if(IsEqualGUID(pMediaType->majortype, MEDIATYPE_Video) && IsEqualGUID(pMediaType->subtype, MEDIASUBTYPE_YUY2) && IsEqualGUID(pMediaType->formattype, FORMAT_VideoInfo))
			{
				_A(pMediaType->pbFormat && pMediaType->cbFormat >= sizeof (VIDEOINFOHEADER));
				const VIDEOINFOHEADER* pVideoInfoHeader = (const VIDEOINFOHEADER*) pMediaType->pbFormat;
				// ASSU: rcSource and rcTarget define whole frame
				const BITMAPINFOHEADER* pBitmapInfoHeader = &pVideoInfoHeader->bmiHeader;
				if(pBitmapInfoHeader->biCompression == FOURCC_YUY2)
				{
					return TRUE;
				}
			}
		}
		return FALSE;
	}
	static BOOL IsValidMediaType(const DMO_MEDIA_TYPE* pMediaType) throw()
	{
		return IsValidMediaType(reinterpret_cast<const CDmoMediaType&>(pMediaType));
	}
	static BYTE Adjust(BYTE nY, LONG nBrightness, LONG nContrast)
	{
		static const INT g_nOffset = 16;
		_A((1 << 14) == 0x4000);
		INT nIntY = ((((INT) nY - g_nOffset) * nContrast) >> 14) + nBrightness + g_nOffset;
		if(nIntY < 0)
			nIntY = 0;
		else if(nIntY > 0xFF)
			nIntY = 0xFF;
		return (BYTE) nIntY;
	}

public:
// CBrightnessContrastObject
	static LPCTSTR GetObjectFriendlyName() throw()
	{
		return _T("Alax.Info BrightnessContrast Sample");
	}
	static VOID RegisterMediaObject()
	{
		DMO_PARTIAL_MEDIATYPE pInputMediaTypes[] = 
		{
			{ MEDIATYPE_Video, MEDIASUBTYPE_YUY2 },
		};
		DMO_PARTIAL_MEDIATYPE pOutputMediaTypes[] = 
		{
			{ MEDIATYPE_Video, MEDIASUBTYPE_YUY2 },
		};
		static const DWORD g_nFlags = 0;
		AtlCheck(DMORegister(CStringW(GetObjectFriendlyName()), GetObjectCLSID(), DMOCATEGORY_VIDEO_EFFECT, g_nFlags, _countof(pInputMediaTypes), pInputMediaTypes, _countof(pOutputMediaTypes), pOutputMediaTypes));
	}
	static VOID UnregisterMediaObject()
	{
		AtlCheck(DMOUnregister(GetObjectCLSID(), DMOCATEGORY_VIDEO_EFFECT));
	}
	static HRESULT WINAPI UpdateRegistry(BOOL bRegister) throw()
	{
		_ATLTRY
		{
			if(!bRegister)
				UnregisterMediaObject();
			_A(_pAtlModule);
			_pAtlModule->UpdateRegistryFromResource(IDR_BRIGHTNESSCONTRASTOBJECT, bRegister);
			if(bRegister)
				RegisterMediaObject();
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	CBrightnessContrastObject() throw() :
		m_nBrightness(0x0000),
		m_nContrast(0x4000)
	{
		ATLTRACE2(atlTraceRefcount, 4, _T("this 0x%08x\n"), this);
		// WARN: Initial constants until flexible configuration is implemented
		m_nBrightness = -0x0010;
		m_nContrast = 0x2000;
	}
	~CBrightnessContrastObject() throw()
	{
		ATLTRACE2(atlTraceRefcount, 4, _T("this 0x%08x\n"), this);
	}

// CComObjectRootEx
	VOID Lock()
	{
		CComObjectRootEx<CComMultiThreadModel>::Lock();
	}

// IBrightnessContrastObject

// IMediaObject
	STDMETHOD(GetStreamCount)(DWORD* pnInputStreamCount, DWORD* pnOutputStreamCount) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("...\n"));
		_ATLTRY
		{
			__D(pnInputStreamCount || pnOutputStreamCount, E_POINTER);
			if(pnInputStreamCount)
				*pnInputStreamCount = 1;
			if(pnOutputStreamCount)
				*pnOutputStreamCount = 1;
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetInputStreamInfo)(DWORD nInputStreamIndex, DWORD* pnFlags) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d\n"), nInputStreamIndex);
		_ATLTRY
		{
			__D(pnFlags, E_POINTER);
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			*pnFlags = DMO_INPUT_STREAMF_WHOLE_SAMPLES | DMO_INPUT_STREAMF_SINGLE_SAMPLE_PER_BUFFER | DMO_INPUT_STREAMF_FIXED_SAMPLE_SIZE;
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetOutputStreamInfo)(DWORD nOutputStreamIndex, DWORD* pnFlags) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nOutputStreamIndex %d\n"), nOutputStreamIndex);
		_ATLTRY
		{
			__D(pnFlags, E_POINTER);
			__D(nOutputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			*pnFlags = DMO_OUTPUT_STREAMF_WHOLE_SAMPLES | DMO_OUTPUT_STREAMF_SINGLE_SAMPLE_PER_BUFFER | DMO_OUTPUT_STREAMF_FIXED_SAMPLE_SIZE;
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetInputType)(DWORD nInputStreamIndex, DWORD nTypeIndex, DMO_MEDIA_TYPE* pMediaType) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d, nTypeIndex %d\n"), nInputStreamIndex, nTypeIndex);
		_ATLTRY
		{
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			__D(nTypeIndex >= 0, E_INVALIDARG);
			// NOTE: We bascially don't need to enumerate input media types and only check suggested and reject unsupported
			//       However why not make a last chance suggestion
			if(nTypeIndex >= 1)
				return DMO_E_NO_MORE_ITEMS;
			if(pMediaType)
			{
				CSize Extent(320, 240);
				{
					CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
					if(m_pOutputMediaType)
						Extent = GetExtent(m_pOutputMediaType);
				}
				CreateMediaType(Extent).Detach(pMediaType);
			}
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetOutputType)(DWORD nOutputStreamIndex, DWORD nTypeIndex, DMO_MEDIA_TYPE* pMediaType) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nOutputStreamIndex %d, nTypeIndex %d\n"), nOutputStreamIndex, nTypeIndex);
		_ATLTRY
		{
			__D(nOutputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			__D(nTypeIndex >= 0, E_INVALIDARG);
			if(nTypeIndex >= 1)
				return DMO_E_NO_MORE_ITEMS;
			if(pMediaType)
			{
				CSize Extent(320, 240);
				{
					CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
					if(m_pInputMediaType)
						Extent = GetExtent(m_pInputMediaType);
				}
				CreateMediaType(Extent).Detach(pMediaType);
			}
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(SetInputType)(DWORD nInputStreamIndex, const DMO_MEDIA_TYPE* pMediaType, DWORD nFlags) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d, nFlags 0x%x\n"), nInputStreamIndex, nFlags);
		_ATLTRY
		{
			__D(pMediaType || (nFlags & DMO_SET_TYPEF_CLEAR), E_INVALIDARG);
			__D((nFlags & (DMO_SET_TYPEF_TEST_ONLY | DMO_SET_TYPEF_CLEAR)) != (DMO_SET_TYPEF_TEST_ONLY | DMO_SET_TYPEF_CLEAR), E_INVALIDARG);
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			const CDmoMediaType& pInputMediaType = reinterpret_cast<const CDmoMediaType&>(pMediaType);
			BOOL bValid = IsValidMediaType(pInputMediaType);
			if(bValid && m_pOutputMediaType)
				bValid = GetExtent(pInputMediaType) == GetExtent(m_pOutputMediaType);
			if(nFlags & DMO_SET_TYPEF_TEST_ONLY)
				return bValid ? S_OK : S_FALSE;
			__D(!m_Input.IsSet(), E_UNNAMED);
			if(!(nFlags & DMO_SET_TYPEF_CLEAR))
			{
				if(!bValid)
					return DMO_E_TYPE_NOT_ACCEPTED;
				m_pInputMediaType = pMediaType;
			} else
				m_pInputMediaType.Free();
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(SetOutputType)(DWORD nOutputStreamIndex, const DMO_MEDIA_TYPE* pMediaType, DWORD nFlags) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nOutputStreamIndex %d, nFlags 0x%x\n"), nOutputStreamIndex, nFlags);
		_ATLTRY
		{
			__D(pMediaType || (nFlags & DMO_SET_TYPEF_CLEAR), E_INVALIDARG);
			__D((nFlags & (DMO_SET_TYPEF_TEST_ONLY | DMO_SET_TYPEF_CLEAR)) != (DMO_SET_TYPEF_TEST_ONLY | DMO_SET_TYPEF_CLEAR), E_INVALIDARG);
			__D(nOutputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			const CDmoMediaType& pOutputMediaType = reinterpret_cast<const CDmoMediaType&>(pMediaType);
			BOOL bValid = IsValidMediaType(pOutputMediaType);
			if(bValid && m_pInputMediaType)
				bValid = GetExtent(m_pInputMediaType) == GetExtent(pOutputMediaType);
			if(nFlags & DMO_SET_TYPEF_TEST_ONLY)
				return bValid ? S_OK : S_FALSE;
			__D(!m_Input.IsSet(), E_UNNAMED);
			if(!(nFlags & DMO_SET_TYPEF_CLEAR))
			{
				if(!bValid)
					return DMO_E_TYPE_NOT_ACCEPTED;
				m_pOutputMediaType = pMediaType;
			} else
				m_pOutputMediaType.Free();
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetInputCurrentType)(DWORD nInputStreamIndex, DMO_MEDIA_TYPE* pMediaType) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d\n"), nInputStreamIndex);
		_ATLTRY
		{
			__D(pMediaType, E_POINTER);
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			if(!m_pInputMediaType)
				return DMO_E_TYPE_NOT_SET;
			__C(MoCopyMediaType(pMediaType, m_pInputMediaType));
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetOutputCurrentType)(DWORD nOutputStreamIndex, DMO_MEDIA_TYPE* pMediaType) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nOutputStreamIndex %d\n"), nOutputStreamIndex);
		_ATLTRY
		{
			__D(pMediaType, E_POINTER);
			__D(nOutputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			if(!m_pOutputMediaType)
				return DMO_E_TYPE_NOT_SET;
			__C(MoCopyMediaType(pMediaType, m_pOutputMediaType));
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetInputSizeInfo)(DWORD nInputStreamIndex, DWORD* pnBufferSize, DWORD* pnMaximalLookAheadBufferSize, DWORD* pnAlignment) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d\n"), nInputStreamIndex);
		_ATLTRY
		{
			__D(pnBufferSize && pnMaximalLookAheadBufferSize && pnAlignment, E_POINTER);
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			if(!m_pInputMediaType)
				return DMO_E_TYPE_NOT_SET;
			CSize Extent = GetExtent(m_pInputMediaType);
			*pnBufferSize = (DWORD) Extent.cy * (Extent.cx * 2);
			*pnMaximalLookAheadBufferSize = 0;
			*pnAlignment = 1;
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetOutputSizeInfo)(DWORD nOutputStreamIndex, DWORD* pnBufferSize, DWORD* pnAlignment) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nOutputStreamIndex %d\n"), nOutputStreamIndex);
		_ATLTRY
		{
			__D(pnBufferSize && pnAlignment, E_POINTER);
			__D(nOutputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			if(!m_pOutputMediaType)
				return DMO_E_TYPE_NOT_SET;
			CSize Extent = GetExtent(m_pOutputMediaType);
			*pnBufferSize = (DWORD) Extent.cy * (Extent.cx * 2);
			*pnAlignment = 1;
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetInputMaxLatency)(DWORD nInputStreamIndex, REFERENCE_TIME* pnMaximalLatency) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d\n"), nInputStreamIndex);
		//__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
		return E_NOTIMPL;
	}
	STDMETHOD(SetInputMaxLatency)(DWORD nInputStreamIndex, REFERENCE_TIME nMaximalLatency) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d, nMaximalLatency %d\n"), nInputStreamIndex, (ULONG) nMaximalLatency);
		//__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
		return E_NOTIMPL;
	}
	STDMETHOD(Flush)() throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("...\n"));
		_ATLTRY
		{
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			m_Input.Reset();
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(Discontinuity)(DWORD nInputStreamIndex) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d\n"), nInputStreamIndex);
		_ATLTRY
		{
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			_A(!m_Input.IsSet());
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(AllocateStreamingResources)() throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("...\n"));
		_ATLTRY
		{
			// TODO: ...
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(FreeStreamingResources)() throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("...\n"));
		_ATLTRY
		{
			// TODO: ...
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(GetInputStatus)(DWORD nInputStreamIndex, DWORD* pnFlags) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d\n"), nInputStreamIndex);
		_ATLTRY
		{
			__D(pnFlags, E_POINTER);
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			// TODO: ...
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			*pnFlags = (m_pInputMediaType && m_pOutputMediaType && !m_Input.IsSet()) ? DMO_INPUT_STATUSF_ACCEPT_DATA : 0;
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(Lock)(LONG bLock) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("bLock %d\n"), bLock);
		_ATLTRY
		{
			if(bLock)
				m_DataCriticalSection.Lock();
			else
				m_DataCriticalSection.Unlock();
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(ProcessInput)(DWORD nInputStreamIndex, IMediaBuffer* pMediaBuffer, DWORD nFlags, REFERENCE_TIME nTime, REFERENCE_TIME nLength) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nInputStreamIndex %d, nFlags 0x%x\n"), nInputStreamIndex, nFlags);
		_ATLTRY
		{
			__D(nInputStreamIndex == 0, DMO_E_INVALIDSTREAMINDEX);
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			__D(m_pInputMediaType, DMO_E_NOTACCEPTING);
			__D(!m_Input.IsSet(), DMO_E_NOTACCEPTING);
			m_Input.Set(nFlags, pMediaBuffer, nTime, nLength);
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}
	STDMETHOD(ProcessOutput)(DWORD nFlags, DWORD nOutputBufferCount, DMO_OUTPUT_DATA_BUFFER* pOutputBuffers, DWORD* pnStatus) throw()
	{
		ATLTRACE2(atlTraceCOM, 4, _T("nFlags 0x%x, nOutputBufferCount %d\n"), nFlags, nOutputBufferCount);
		_ATLTRY
		{
			__D(nOutputBufferCount == 1, E_INVALIDARG);
			__D(pOutputBuffers, E_INVALIDARG);
			__D(pnStatus, E_POINTER);
			DMO_OUTPUT_DATA_BUFFER& OutputBuffer = pOutputBuffers[0];
			OutputBuffer.dwStatus = 0;
			*pnStatus = 0;
			CComCritSecLock<CComAutoCriticalSection> DataLock(m_DataCriticalSection);
			if(!m_Input.IsSet())
				return S_FALSE;
			const CSize Extent = GetExtent(m_pInputMediaType);
			_A(Extent == GetExtent(m_pOutputMediaType));
			CInputMediaBuffer InputMediaBuffer(m_Input);
			COutputMediaBuffer OutputMediaBuffer(OutputBuffer);
			_A(m_Input.m_nFlags & DMO_INPUT_DATA_BUFFERF_SYNCPOINT);
			OutputBuffer.dwStatus |= DMO_OUTPUT_DATA_BUFFERF_SYNCPOINT;
			if(m_Input.m_nFlags & DMO_INPUT_DATA_BUFFERF_TIME)
			{
				OutputBuffer.dwStatus |= DMO_OUTPUT_DATA_BUFFERF_TIME;
				OutputBuffer.rtTimestamp = m_Input.m_nTime;
			}
			if(m_Input.m_nFlags & DMO_INPUT_DATA_BUFFERF_TIMELENGTH)
			{
				OutputBuffer.dwStatus |= DMO_OUTPUT_DATA_BUFFERF_TIMELENGTH;
				OutputBuffer.rtTimelength = m_Input.m_nLength;
			}
			__C(AllocateStreamingResources());
			_A(!InputMediaBuffer.m_nDataPosition && !OutputMediaBuffer.m_nDataPosition);
			const BYTE* pnInputData1 = InputMediaBuffer.m_pnData;
			BYTE* pnOutputData1 = OutputMediaBuffer.m_pnData;
			const VIDEOINFOHEADER* pInputVideoInfoHeader = (const VIDEOINFOHEADER*) m_pInputMediaType->pbFormat;
			const VIDEOINFOHEADER* pOutputVideoInfoHeader = (const VIDEOINFOHEADER*) m_pOutputMediaType->pbFormat;
			// NOTE: Output stride may be affected by video renderer optimization
			const SSIZE_T nInputStride = pInputVideoInfoHeader->bmiHeader.biWidth * 2;
			const SSIZE_T nOutputStride = pOutputVideoInfoHeader->bmiHeader.biWidth * 2;
			_A(nInputStride >= Extent.cx * 2 && nOutputStride >= Extent.cx * 2);
			_A(InputMediaBuffer.m_nDataLength >= (DWORD) (Extent.cy * nInputStride));
			_A(OutputMediaBuffer.m_nDataLength >= (DWORD) (Extent.cy * nOutputStride));
			//_A(pInputVideoInfoHeader->bmiHeader.biHeight == pOutputVideoInfoHeader->bmiHeader.biHeight);
			const LONG& nBrightness = m_nBrightness;
			const LONG& nContrast = m_nContrast;
			// SUGG: Optimizible using SIMD CPU instructions and Intel Performance Primitives
			// NOTE: Per-macropixel processing
			for(LONG nRow = Extent.cy; nRow > 0; nRow--)
			{
				const BYTE* pnInputData2 = pnInputData1;
				BYTE* pnOutputData2 = pnOutputData1;
				for(LONG nMacroColumn = Extent.cx / 2; nMacroColumn > 0; nMacroColumn--)
				{
					pnOutputData2[0] = Adjust(pnInputData2[0], nBrightness, nContrast);
					pnOutputData2[1] = pnInputData2[1];
					pnOutputData2[2] = Adjust(pnInputData2[2], nBrightness, nContrast);
					pnOutputData2[3] = pnInputData2[3];
					pnInputData2 += 4;
					pnOutputData2 += 4;
				}
				pnInputData1 += nInputStride;
				pnOutputData1 += nOutputStride;
			}
			_A(pnInputData1 - InputMediaBuffer.m_pnData == Extent.cy * nInputStride);
			_A(pnOutputData1 - OutputMediaBuffer.m_pnData == Extent.cy * nOutputStride);
			InputMediaBuffer.m_nDataPosition = Extent.cy * nInputStride;
			OutputBuffer.pBuffer->SetLength(Extent.cy * nOutputStride);
			m_Input.Reset();
		}
		_ATLCATCH(Exception)
		{
			return Exception;
		}
		return S_OK;
	}

};

OBJECT_ENTRY_AUTO(__uuidof(BrightnessContrastObject), CBrightnessContrastObject)
