////////////////////////////////////////////////////////////
// rodmo.h - 
//
// Copyright (C) Alax.Info, 2006-2008
// http://alax.info
//
// A permission to use the source code is granted as long as reference to 
// source website http://alax.info is retained.

#pragma once

#include <dmo.h>
#include <dmodshow.h>

#pragma comment(lib, "dmoguids.lib")
#pragma comment(lib, "msdmo.lib")

////////////////////////////////////////////////////////////
// CDmoMediaTypeT, CDmoMediaType

template <BOOL t_bManaged>
class CDmoMediaTypeT
{
private:
	DMO_MEDIA_TYPE* m_pMediaType;

public:
// CDmoMediaTypeT
	CDmoMediaTypeT() throw()
	{
		if(t_bManaged)
			m_pMediaType = NULL;
	}
	CDmoMediaTypeT(const DMO_MEDIA_TYPE* pMediaType)
	{
		if(t_bManaged && pMediaType)
		{
			__C(MoDuplicateMediaType(&m_pMediaType, pMediaType));
		} else
			m_pMediaType = const_cast<DMO_MEDIA_TYPE*>(pMediaType);
	}
	CDmoMediaTypeT(CDmoMediaTypeT& MediaType) throw()
	{
		if(t_bManaged && MediaType.m_pMediaType)
		{
			// NOTE: Transfering ownership
			m_pMediaType = MediaType.m_pMediaType;
			MediaType.m_pMediaType = NULL;
		} else
			m_pMediaType = MediaType.m_pMediaType;
	}
	~CDmoMediaTypeT() throw()
	{
		if(t_bManaged)
			Free();
	}
	operator DMO_MEDIA_TYPE* () const throw()
	{
		return m_pMediaType;
	}
	DMO_MEDIA_TYPE* operator -> () const throw()
	{
		_A(m_pMediaType);
		return m_pMediaType;
	}
	DMO_MEDIA_TYPE** operator & () throw()
	{
		_A(!m_pMediaType):
		return &m_pMediaType;
	}
	CDmoMediaTypeT& operator = (const DMO_MEDIA_TYPE* pMediaType)
	{
		if(t_bManaged && pMediaType)
		{
			Free();
			__C(MoDuplicateMediaType(&m_pMediaType, pMediaType));
		} else
			m_pMediaType = const_cast<DMO_MEDIA_TYPE*>(pMediaType);
		return *this;
	}
	CDmoMediaTypeT& Attach(DMO_MEDIA_TYPE* pMediaType) throw()
	{
		if(t_bManaged)
			Free();
		m_pMediaType = pMediaType;
		return *this;
	}
	DMO_MEDIA_TYPE* Detach() throw()
	{
		DMO_MEDIA_TYPE* pMediaType = m_pMediaType;
		m_pMediaType = NULL;
		return pMediaType;
	}
	VOID Detach(DMO_MEDIA_TYPE* pMediaType) throw()
	{
		_A(!IsBadWritePtr(pMediaType, sizeof (DMO_MEDIA_TYPE)));
		_A(m_pMediaType);
		CopyMemory(pMediaType, m_pMediaType, sizeof (DMO_MEDIA_TYPE));
		ZeroMemory(m_pMediaType, sizeof (DMO_MEDIA_TYPE));
		if(t_bManaged)
			Free();
	}
	VOID Allocate(SIZE_T nFormatSize = 0)
	{
		if(t_bManaged)
			Free();
		__C(MoCreateMediaType(&m_pMediaType, (DWORD) nFormatSize));
	}
	VOID Free() throw()
	{
		if(m_pMediaType)
		{
			_V(MoDeleteMediaType(m_pMediaType));
			m_pMediaType = NULL;
		}
	}
	__if_exists(CMediaType)
	{
		const CMediaType& GetMediaType() const throw()
		{
			return reinterpret_cast<const CMediaType&>(*this);
		}
		CMediaType& GetMediaType() throw()
		{
			return reinterpret_cast<CMediaType&>(*this);
		}
		const CDmoMediaTypeT& AllocateVideoInfo(const GUID& Subtype, LONG nWidth, LONG nHeight, WORD nBitCount = 24, DWORD nCompression = BI_RGB, REFERENCE_TIME nAverageTimePerFrame = 0)
		{
			// HOTFIX: CMediaTypeT's implementation in CDmoMediaTypeT
			CMediaType pMediaType;
			pMediaType.AllocateVideoInfo(Subtype, nWidth, nHeight, nBitCount, nCompression, nAverageTimePerFrame);
			*this = (const DMO_MEDIA_TYPE*) (const AM_MEDIA_TYPE*) pMediaType;
			return *this;
		}
		//const CMediaTypeT& AllocateVideoInfo(const GUID& Subtype, CSize Extent, WORD nBitCount = 24, DWORD nCompression = BI_RGB, REFERENCE_TIME nAverageTimePerFrame = 0)
		const CDmoMediaTypeT& AllocateVideoInfo(LONG nWidth, LONG nHeight, WORD nBitCount = 24, DWORD nCompression = BI_RGB, REFERENCE_TIME nAverageTimePerFrame = 0)
		{
			// HOTFIX: CMediaTypeT's implementation in CDmoMediaTypeT
			CMediaType pMediaType;
			pMediaType.AllocateVideoInfo(nWidth, nHeight, nBitCount, nCompression, nAverageTimePerFrame);
			*this = (const DMO_MEDIA_TYPE*) (const AM_MEDIA_TYPE*) pMediaType;
			return *this;
		}
		//const CMediaTypeT& AllocateVideoInfo(CSize Extent, WORD nBitCount = 24, DWORD nCompression = BI_RGB, REFERENCE_TIME nAverageTimePerFrame = 0)
	}
};

typedef CDmoMediaTypeT<TRUE> CDmoMediaType;

////////////////////////////////////////////////////////////
// CMediaObjectT

//template <typename T>
class ATL_NO_VTABLE CMediaObjectT :
	public IMediaObject
{
public:

	////////////////////////////////////////////////////////
	// CInput

	class CInput
	{
	public:
		DWORD m_nFlags;
		CComPtr<IMediaBuffer> m_pMediaBuffer;
		REFERENCE_TIME m_nTime;
		REFERENCE_TIME m_nLength;
		ULONG m_nPosition;

	public:
	// CInput
		BOOL IsSet() const throw()
		{
			return m_pMediaBuffer != NULL;
		}
		VOID Set(DWORD nFlags, IMediaBuffer* pMediaBuffer, REFERENCE_TIME nTime, REFERENCE_TIME nLength) throw()
		{
			m_nFlags = nFlags;
			m_pMediaBuffer = pMediaBuffer;
			m_nTime = nTime;
			m_nLength = nLength;
			m_nPosition = 0;
		}
		VOID Reset() throw()
		{
			m_pMediaBuffer = NULL;
		}
	};

	////////////////////////////////////////////////////////
	// CInputMediaBuffer

	class CInputMediaBuffer
	{
	public:
		CComPtr<IMediaBuffer>& m_pMediaBuffer;		// Data container reference
		const BYTE* m_pnData;						// Data pointer
		DWORD m_nDataLength;						// Data length, bytes
		DWORD& m_nDataPosition;						// Data position, bytes
		DWORD m_nFrameLength;						// Frame length, bytes

	public:
	// CInputMediaBuffer
		CInputMediaBuffer(CInput& Input, DWORD nFrameLength = 0) :
			m_pMediaBuffer(Input.m_pMediaBuffer),
			m_pnData(NULL),
			m_nDataLength(0),
			m_nDataPosition(Input.m_nPosition),
			m_nFrameLength(nFrameLength)
		{
			_A(m_pMediaBuffer);
			__C(m_pMediaBuffer->GetBufferAndLength(const_cast<BYTE**>(&m_pnData), &m_nDataLength));
			_Z5(atlTraceGeneral, 5, _T("m_pnData 0x%08x, m_nDataLength %d, m_nDataPosition %d, m_nFrameLength %d\n"), m_pnData, m_nDataLength, m_nDataPosition, m_nFrameLength);
			__D(m_nDataPosition <= m_nDataLength, E_UNNAMED);
		}
	};

	////////////////////////////////////////////////////////
	// COutputMediaBuffer

	class COutputMediaBuffer
	{
	public:
		CComPtr<IMediaBuffer>& m_pMediaBuffer;
		BYTE* m_pnData;
		DWORD m_nDataLength;
		DWORD m_nDataPosition;
		DWORD m_nFrameLength;

	public:
	// COutputMediaBuffer
		COutputMediaBuffer(DMO_OUTPUT_DATA_BUFFER& OutputBuffer, DWORD nFrameLength = 0) :
			m_pMediaBuffer(reinterpret_cast<CComPtr<IMediaBuffer>&>(OutputBuffer.pBuffer)),
			m_pnData(NULL),
			m_nDataLength(0),
			m_nDataPosition(0),
			m_nFrameLength(nFrameLength)
		{
			_A(m_pMediaBuffer);
			DWORD nLength;
			__C(m_pMediaBuffer->GetBufferAndLength(&m_pnData, &nLength));
			_A(!nLength);
			__C(m_pMediaBuffer->GetMaxLength(&m_nDataLength));
			_Z5(atlTraceGeneral, 5, _T("m_pnData 0x%08x, m_nDataLength %d, m_nDataPosition %d, m_nFrameLength %d\n"), m_pnData, m_nDataLength, m_nDataPosition, m_nFrameLength);
		}
	};

public:
// CMediaObjectT

// IMediaObject
	STDMETHOD(Flush)() throw()
	{
		_Z4(atlTraceCOM, 4, _T("...\n"));
		return S_OK;
	}
	STDMETHOD(AllocateStreamingResources)() throw()
	{
		_Z4(atlTraceCOM, 4, _T("...\n"));
		return S_OK;
	}
	STDMETHOD(FreeStreamingResources)() throw()
	{
		_Z4(atlTraceCOM, 4, _T("...\n"));
		return S_OK;
	}
};

////////////////////////////////////////////////////////////
// ConstructDmoWrappedFilter

inline CComPtr<IDMOWrapperFilter> ConstructDmoWrappedFilter(const CLSID& ClassIdentifier, const GUID& Category)
{
	CComPtr<IDMOWrapperFilter> pDmoWrapperFilter;
	__C(pDmoWrapperFilter.CoCreateInstance(CLSID_DMOWrapperFilter));
	__C(pDmoWrapperFilter->Init(ClassIdentifier, Category));
	return pDmoWrapperFilter;
}

